/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import org.xmloperator.lambda3.exception.UndefinedEliminationException;
import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Function;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.model.Variable;

abstract class ConstantImpl<F extends Function> extends TermImpl implements
		Constant {
	F function = null;
	int argumentCount = 0;

	ConstantImpl() {
	}

	/**
	 * Replaces an argument.
	 * 
	 * @param argumentIndex
	 *            the index of the argument.
	 * @param newValue
	 *            a new value for the given argument.
	 */
	abstract void replaceArgument(int argumentIndex, TermImpl newValue);

	// ////////////////////////
	// Constant implementation
	// ////////////////////////

	public final String getCombinatorName() {
		return this.function.getName();
	}

	public final boolean acceptArgument(Term arg) {
		return this.function.acceptArgument(this.argumentCount, arg);
	}

	public final int getArgumentCount() {
		return this.argumentCount;
	}

	public boolean isInt() {
		return false;
	}

	public boolean isFloat() {
		return false;
	}

	// ////////////////////
	// Term implementation
	// ////////////////////

	public final boolean isConstant() {
		return true;
	}

	public final boolean isBroken() {
		return this.function == null;
	}

	public final Term abstractionElimination(TermRecycler termRecycler) {
		throw new UndefinedEliminationException();
	}

	// //////////////////
	// TermImpl override
	// //////////////////

	final TermImpl abstractionElimination(Variable boundVariable,
			TermRecycler termRecycler) {
		throw new UndefinedEliminationException();
	}
}
