/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.ArrayList;

import org.xmloperator.lambda3.model.Constant;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermRecycler;
import org.xmloperator.lambda3.util.ReservableCollection;

/**
 * TermRecycler implementation.
 */
public class TermRecyclerImpl implements TermRecycler {
	final ReservableCollection<ArrayList<TermImpl>> reservableTermStack = new ReservableCollection<ArrayList<TermImpl>>(
			new ArrayList<TermImpl>());
	final ReservableCollection<ArrayList<VariableImpl>> reservableVariableList = new ReservableCollection<ArrayList<VariableImpl>>(
			new ArrayList<VariableImpl>());

	// ////////////////////////////
	// TermRecycler implementation
	// ////////////////////////////

	public void breakTermDeeply(Term rootTerm) {
		if (rootTerm == null)
			throw new IllegalArgumentException();
		ArrayList<TermImpl> termStack = this.reservableTermStack.reserve();
		ArrayList<VariableImpl> variableList = this.reservableVariableList
				.reserve();
		TermImpl term = (TermImpl) rootTerm;
		try {
			while (true) {
				if (term.isVariable() || term.referenceCount != 0
						|| term.isBroken()) {
					if (termStack.isEmpty()) {
						for (VariableImpl variable : variableList)
							variable.breakIt();
						return;
					}
					int index = termStack.size() - 1;
					term = termStack.remove(index);
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					variableList.add(abstraction.boundVariable);
					term = abstraction.body;
					abstraction.breakIt();
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					termStack.add(application.argument);
					term = application.function;
					application.breakIt();
				} else if (term.isConstant()) {
					Constant constant = (Constant) term;
					TermImpl arg = (TermImpl) constant.removeLastArg();
					if (arg == null) {
						((TermImpl) constant).breakIt();
						if (termStack.isEmpty()) {
							for (VariableImpl variable : variableList)
								variable.breakIt();
							return;
						}
						int index = termStack.size() - 1;
						term = termStack.remove(index);
					} else {
						term = arg;
						while (true) {
							arg = (TermImpl) constant.removeLastArg();
							if (arg == null)
								break;
							termStack.add(arg);
						}
						((TermImpl) constant).breakIt();
					}
				} else
					throw new IllegalStateException();
			}
		} finally {
			this.reservableTermStack.makeBack();
			this.reservableVariableList.makeBack();
		}
	}
}
