/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import org.xmloperator.lambda3.exception.AssertionContradictedException;
import org.xmloperator.lambda3.exception.term.BrokenTermException;
import org.xmloperator.lambda3.exception.term.NotARedexException;
import org.xmloperator.lambda3.exception.term.RedexSharingException;
import org.xmloperator.lambda3.exception.term.StillReferencedTermException;
import org.xmloperator.lambda3.exception.term.TermException;
import org.xmloperator.lambda3.model.Application;
import org.xmloperator.lambda3.model.InformantReduction;
import org.xmloperator.lambda3.model.Term;
import org.xmloperator.lambda3.model.TermReducer;
import org.xmloperator.lambda3.util.ReservableCollection;
import org.xmloperator.lambda3.util.ReservableMap;

/**
 * TermReducer implementation.
 */
public class TermReducerImpl extends TermRecyclerImpl implements TermReducer {
	static final int I_RETURN = 0;
	/**
	 * Abstraction body, application function or constant first argument.
	 */
	static final int I_FUNCTION = 0;
	/**
	 * Application argument or constant second argument.
	 */
	static final int I_ARGUMENT = 1;

	static final Integer ZERO = Integer.valueOf(0);
	static final Integer ONE = Integer.valueOf(1);
	static final Integer IND_RETURN = ZERO;
	static final Integer IND_FUNCTION = ZERO;
	static final Integer IND_ARGUMENT = ONE;

	static final Integer toInteger(int n) {
		return n == 1 ? ONE : Integer.valueOf(n);
	}

	final ReservableCollection<ArrayList<Integer>> reservableIntegerStack = new ReservableCollection<ArrayList<Integer>>(
			new ArrayList<Integer>());
	final ReservableCollection<HashSet<Term>> reservableExplorationSet = new ReservableCollection<HashSet<Term>>(
			new HashSet<Term>());
	final ReservableMap<HashMap<Term, TermImpl>> reservableDuplicationMap = new ReservableMap<HashMap<Term, TermImpl>>(
			new HashMap<Term, TermImpl>());
	final ReservableCollection<ArrayList<TermImpl>> reservableTermList = new ReservableCollection<ArrayList<TermImpl>>(
			new ArrayList<TermImpl>());
	final ReservableCollection<ArrayList<Integer>> reservableIntegerList = new ReservableCollection<ArrayList<Integer>>(
			new ArrayList<Integer>());

	/**
	 * Replaces within a root Term each reference to an "old" Term (or none) by
	 * a reference to a "new" Term (or none).
	 * 
	 * @param rootTerm
	 *            the root Term.
	 * @param referenceCount
	 *            the old Term reference count.
	 * @param oldTerm
	 *            the old Term.
	 * @param newTerm
	 *            the new Term.
	 */
	final void substitude(final TermImpl rootTerm, final int referenceCount,
			final Term oldTerm, final TermImpl newTerm) {
		if (rootTerm == null)
			throw new IllegalArgumentException();
		ArrayList<TermImpl> termStack = this.reservableTermStack.reserve();
		ArrayList<Integer> todoArgumentStack = this.reservableIntegerStack
				.reserve();
		HashSet<Term> explorationSet = this.reservableExplorationSet.reserve();
		int count = 0;
		TermImpl term = rootTerm;
		int todoArgument = I_FUNCTION;
		boolean onReturn = false;
		try {
			while (true) {
				if (onReturn
						|| term.isVariable()
						|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)
						|| (term.referenceCount > 1 && explorationSet
								.contains(term))) {
					if (term == oldTerm)
						throw new AssertionContradictedException();
					if (termStack.isEmpty())
						return;
					if (onReturn && term.referenceCount > 1)
						explorationSet.add(term);
					int index = termStack.size() - 1;
					term = termStack.remove(index);
					todoArgument = todoArgumentStack.remove(index).intValue();
					onReturn = todoArgument == I_RETURN;
				} else if (term.isConstant()) {
					ConstantImpl<?> constant = (ConstantImpl<?>) term;
					boolean isLastArgument = todoArgument == constant
							.getArgumentCount() - 1;
					if (constant.getArgument(todoArgument) == oldTerm) {
						constant.replaceArgument(todoArgument, newTerm);
						count++;
						if (count == referenceCount)
							return;
						if (isLastArgument) {
							if (termStack.isEmpty())
								return;
							int index = termStack.size() - 1;
							term = termStack.remove(index);
							todoArgument = todoArgumentStack.remove(index)
									.intValue();
							onReturn = todoArgument == I_RETURN;
						} else {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							todoArgument++;
						}
					} else {
						if (isLastArgument) {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
						} else {
							termStack.add(term);
							todoArgumentStack.add(toInteger(todoArgument + 1));
						}
						term = (TermImpl) constant.getArgument(todoArgument);
						if (term == null)
							throw new BrokenTermException(rootTerm, constant);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					if (abstraction.body == oldTerm) {
						abstraction.replaceBody(newTerm);
						count++;
						if (count == referenceCount || termStack.isEmpty())
							return;
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else {
						if (term.referenceCount > 1) {
							termStack.add(term);
							todoArgumentStack.add(IND_RETURN);
						}
						term = abstraction.body;
						if (term == null)
							throw new BrokenTermException(rootTerm, abstraction);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (todoArgument == I_ARGUMENT) {
						if (application.argument == oldTerm) {
							application.replaceArgument(newTerm);
							count++;
							if (count == referenceCount || termStack.isEmpty())
								return;
							int index = termStack.size() - 1;
							term = termStack.remove(index);
							todoArgument = todoArgumentStack.remove(index)
									.intValue();
							onReturn = todoArgument == I_RETURN;
						} else {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							term = application.argument;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = I_FUNCTION;
						}
					} else {
						if (application.function == oldTerm) {
							application.replaceFunction(newTerm);
							count++;
							if (count == referenceCount)
								return;
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							todoArgument = I_ARGUMENT;
						} else {
							termStack.add(term);
							todoArgumentStack.add(IND_ARGUMENT);
							term = application.function;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = I_FUNCTION;
						}
					}
				} else
					throw new IllegalStateException();
			}
		} finally {
			this.reservableTermStack.makeBack();
			this.reservableIntegerStack.makeBack();
			this.reservableExplorationSet.makeBack();
		}
	}

	/**
	 * Locates within a root Term each reference to a given Term.
	 * 
	 * @param rootTerm
	 *            the root Term.
	 * @param locatedTerm
	 *            the Term to locate.
	 * @param referenceCount
	 *            the expected location count.
	 * @param parentTermList
	 *            the resulting List of parent terms.
	 * @param argumentIndexList
	 *            the resulting List of argument indicator.
	 */
	final void locate(final TermImpl rootTerm, final Term locatedTerm,
			final int referenceCount, final ArrayList<TermImpl> parentTermList,
			final ArrayList<Integer> argumentIndexList) {
		if (rootTerm == null || !parentTermList.isEmpty()
				|| !argumentIndexList.isEmpty())
			throw new IllegalArgumentException();
		ArrayList<TermImpl> termStack = this.reservableTermStack.reserve();
		ArrayList<Integer> todoArgumentStack = this.reservableIntegerStack
				.reserve();
		HashSet<Term> explorationSet = this.reservableExplorationSet.reserve();
		int count = 0;
		TermImpl term = rootTerm;
		int todoArgument = I_FUNCTION;
		boolean onReturn = false;
		try {
			while (true) {
				if (onReturn
						|| term.isVariable()
						|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)
						|| (term.referenceCount > 1 && explorationSet
								.contains(term))) {
					if (term == locatedTerm)
						throw new AssertionContradictedException();
					if (termStack.isEmpty())
						return;
					if (onReturn && term.referenceCount > 1)
						explorationSet.add(term);
					int index = termStack.size() - 1;
					term = termStack.remove(index);
					todoArgument = todoArgumentStack.remove(index).intValue();
					onReturn = todoArgument == I_RETURN;
				} else if (term.isConstant()) {
					ConstantImpl<?> constant = (ConstantImpl<?>) term;
					boolean isLastArgument = todoArgument == constant
							.getArgumentCount() - 1;
					if (constant.getArgument(todoArgument) == locatedTerm) {
						parentTermList.add(constant);
						argumentIndexList.add(toInteger(todoArgument));
						count++;
						if (count == referenceCount)
							return;
						if (isLastArgument) {
							if (termStack.isEmpty())
								return;
							int index = termStack.size() - 1;
							term = termStack.remove(index);
							todoArgument = todoArgumentStack.remove(index)
									.intValue();
							onReturn = todoArgument == I_RETURN;
						} else {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							todoArgument++;
						}
					} else {
						if (isLastArgument) {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
						} else {
							termStack.add(term);
							todoArgumentStack.add(toInteger(todoArgument + 1));
						}
						term = (TermImpl) constant.getArgument(todoArgument);
						if (term == null)
							throw new BrokenTermException(rootTerm, constant);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					if (abstraction.body == locatedTerm) {
						parentTermList.add(abstraction);
						argumentIndexList.add(IND_FUNCTION);
						count++;
						if (count == referenceCount || termStack.isEmpty())
							return;
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else {
						if (term.referenceCount > 1) {
							termStack.add(term);
							todoArgumentStack.add(IND_RETURN);
						}
						term = abstraction.body;
						if (term == null)
							throw new BrokenTermException(rootTerm, abstraction);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (todoArgument == I_ARGUMENT) {
						if (application.argument == locatedTerm) {
							parentTermList.add(application);
							argumentIndexList.add(IND_ARGUMENT);
							count++;
							if (count == referenceCount || termStack.isEmpty())
								return;
							int index = termStack.size() - 1;
							term = termStack.remove(index);
							todoArgument = todoArgumentStack.remove(index)
									.intValue();
							onReturn = todoArgument == I_RETURN;
						} else {
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							term = application.argument;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = I_FUNCTION;
						}
					} else {
						if (application.function == locatedTerm) {
							parentTermList.add(application);
							argumentIndexList.add(IND_FUNCTION);
							count++;
							if (count == referenceCount)
								return;
							if (term.referenceCount > 1) {
								termStack.add(term);
								todoArgumentStack.add(IND_RETURN);
							}
							todoArgument = I_ARGUMENT;
						} else {
							termStack.add(term);
							todoArgumentStack.add(IND_ARGUMENT);
							term = application.function;
							if (term == null)
								throw new BrokenTermException(rootTerm,
										application);
							todoArgument = I_FUNCTION;
						}
					}
				} else
					throw new IllegalStateException();
			}
		} finally {
			this.reservableTermStack.makeBack();
			this.reservableIntegerStack.makeBack();
			this.reservableExplorationSet.makeBack();
		}
	}

	/**
	 * Replaces each located instance of an old Term by a new Term.
	 * 
	 * @param parentTermList
	 *            a List of parent terms.
	 * @param argumentIndexList
	 *            a List of argument indicator.
	 * @param oldTerm
	 *            the old Term.
	 * @param newTerm
	 *            the new Term.
	 */
	static final void replace(final ArrayList<TermImpl> parentTermList,
			final ArrayList<Integer> argumentIndexList, Term oldTerm,
			TermImpl newTerm) {
		if (parentTermList.size() != argumentIndexList.size())
			throw new IllegalArgumentException();
		for (int k = 0; k < parentTermList.size(); k++) {
			Term parentTerm = parentTermList.get(k);
			if (parentTerm.isAbstraction()) {
				AbstractionImpl abstraction = (AbstractionImpl) parentTerm;
				if (abstraction.body != oldTerm)
					throw new IllegalStateException();
				abstraction.replaceBody(newTerm);
			} else if (parentTerm.isApplication()) {
				ApplicationImpl application = (ApplicationImpl) parentTerm;
				boolean isArgument = (argumentIndexList.get(k) == IND_ARGUMENT);
				if (isArgument) {
					if (application.argument != oldTerm)
						throw new IllegalStateException();
					application.replaceArgument(newTerm);
				} else {
					if (application.function != oldTerm)
						throw new IllegalStateException();
					application.replaceFunction(newTerm);
				}
			} else if (parentTerm.isConstant()) {
				ConstantImpl<?> constant = (ConstantImpl<?>) parentTerm;
				int argumentIndex = argumentIndexList.get(k);
				if (constant.getArgument(argumentIndex) != oldTerm)
					throw new IllegalStateException();
				constant.replaceArgument(argumentIndex, newTerm);
			} else
				throw new IllegalStateException();
		}
	}

	/**
	 * Creates a duplication of a root Term and returns it.
	 * <p>
	 * The closed abstractions are not duplicated (root excepted).
	 * </p>
	 * 
	 * @param rootTerm
	 *            a source Term.
	 * @return a minimal duplication of the given root Term.
	 */
	final TermImpl duplicate(final Term rootTerm) {
		if (rootTerm == null)
			throw new IllegalArgumentException();
		ArrayList<TermImpl> termStack = this.reservableTermStack.reserve();
		ArrayList<Integer> todoArgumentStack = this.reservableIntegerStack
				.reserve();
		HashMap<Term, TermImpl> duplicationMap = this.reservableDuplicationMap
				.reserve();
		TermImpl term = (TermImpl) rootTerm;
		int todoArgument = I_FUNCTION;
		boolean onReturn = false;
		try {
			while (true) {
				if (term.isVariable()
						|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)
						|| (term.referenceCount > 1 && duplicationMap
								.containsKey(term))) {
					if (termStack.isEmpty()) {
						if (term.isConstant()
								&& ((ConstantImpl<?>) term).argumentCount == 0)
							return (TermImpl) ((ConstantImpl<?>) term)
									.cloneWithoutArgument();
						throw new AssertionContradictedException();
					}
					int index = termStack.size() - 1;
					term = termStack.remove(index);
					todoArgument = todoArgumentStack.remove(index).intValue();
					onReturn = todoArgument == I_RETURN;
				} else if (term.isConstant()) {
					ConstantImpl<?> constant = (ConstantImpl<?>) term;
					if (onReturn) {
						ConstantImpl<?> cloneConstant = (ConstantImpl<?>) constant
								.cloneWithoutArgument();
						for (int k = 0; k < constant.getArgumentCount(); k++) {
							TermImpl argument = (TermImpl) constant
									.getArgument(k);
							TermImpl cloneArgument = duplicationMap
									.get(argument);
							if (cloneArgument == null)
								cloneArgument = argument;
							cloneConstant.apply(cloneArgument, this);
							if (cloneArgument != argument
									&& !cloneArgument.isVariable()
									&& argument.referenceCount == 1)
								duplicationMap.remove(argument);
						}
						if (termStack.isEmpty())
							return cloneConstant;
						duplicationMap.put(constant, cloneConstant);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else {
						termStack.add(term);
						todoArgumentStack.add(todoArgument == constant
								.getArgumentCount() - 1 ? IND_RETURN
								: toInteger(todoArgument + 1));
						term = (TermImpl) constant.getArgument(todoArgument);
						if (term == null)
							throw new BrokenTermException(rootTerm, constant);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isAbstraction()) {
					AbstractionImpl abstraction = (AbstractionImpl) term;
					if (onReturn) {
						VariableImpl cloneVariable = (VariableImpl) duplicationMap
								.get(abstraction.boundVariable);
						if (cloneVariable == null)
							cloneVariable = abstraction.boundVariable;
						TermImpl cloneBody = duplicationMap
								.get(abstraction.body);
						if (cloneBody == null)
							cloneBody = abstraction.body;
						AbstractionImpl cloneAbstraction = (cloneVariable != abstraction.boundVariable) ? AbstractionImpl
								.newAbstraction(cloneVariable, cloneBody)
								: abstraction;
						cloneAbstraction.setClosedLike(abstraction);
						if (cloneVariable != abstraction.boundVariable)
							duplicationMap.remove(abstraction.boundVariable);
						if (cloneBody != abstraction.body
								&& !cloneBody.isVariable()
								&& abstraction.body.referenceCount == 1)
							duplicationMap.remove(abstraction.body);
						if (termStack.isEmpty())
							return cloneAbstraction;
						if (cloneAbstraction != abstraction)
							duplicationMap.put(abstraction, cloneAbstraction);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else if (abstraction != rootTerm
							&& abstraction.isClosed()) {
						if (termStack.isEmpty())
							throw new AssertionContradictedException();
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else {
						VariableImpl cloneVariable = VariableImpl.newVariable();
						duplicationMap.put(abstraction.boundVariable,
								cloneVariable);
						termStack.add(term);
						todoArgumentStack.add(IND_RETURN);
						term = abstraction.body;
						if (term == null)
							throw new BrokenTermException(rootTerm, abstraction);
						todoArgument = I_FUNCTION;
					}
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (onReturn) {
						TermImpl cloneFunction = duplicationMap
								.get(application.function);
						if (cloneFunction == null)
							cloneFunction = application.function;
						TermImpl cloneArgument = duplicationMap
								.get(application.argument);
						if (cloneArgument == null)
							cloneArgument = application.argument;
						TermImpl cloneApplication = (cloneFunction != application.function || cloneArgument != application.argument) ? ApplicationImpl
								.newApplication(cloneFunction, cloneArgument)
								: application;
						if (cloneFunction != application.function
								&& !cloneFunction.isVariable()
								&& application.function.referenceCount == 1)
							duplicationMap.remove(application.function);
						if (cloneArgument != application.argument
								&& !cloneArgument.isVariable()
								&& application.argument.referenceCount == 1)
							duplicationMap.remove(application.argument);
						if (termStack.isEmpty())
							return cloneApplication;
						if (cloneApplication != application)
							duplicationMap.put(application, cloneApplication);
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						onReturn = todoArgument == I_RETURN;
					} else if (todoArgument == I_ARGUMENT) {
						termStack.add(term);
						todoArgumentStack.add(IND_RETURN);
						term = application.argument;
						if (term == null)
							throw new BrokenTermException(rootTerm, application);
						todoArgument = I_FUNCTION;
					} else {
						termStack.add(term);
						todoArgumentStack.add(IND_ARGUMENT);
						term = application.function;
						if (term == null)
							throw new BrokenTermException(rootTerm, application);
						todoArgument = I_FUNCTION;
					}
				} else
					throw new IllegalStateException();
			}
		} finally {
			this.reservableTermStack.makeBack();
			this.reservableIntegerStack.makeBack();
			this.reservableDuplicationMap.makeBack();
		}
	}

	// ///////////////////////////
	// TermReducer implementation
	// ///////////////////////////

	public Term betaDeltaReduce(Application redex) {
		if (!redex.isRedex())
			throw new NotARedexException(redex);
		ApplicationImpl application = (ApplicationImpl) redex;
		if (TermImpl.ASSERTION && application.referenceCount > 0)
			throw new StillReferencedTermException(application);
		if (application.function.referenceCount > 1) {
			TermImpl cloneFunction = this.duplicate(application.function);
			if (cloneFunction != application.function)
				application.replaceFunction(cloneFunction);
		}
		TermImpl argument = application.argument;
		if (application.function.isConstant()) {
			// Delta reduction
			ConstantImpl<?> constant = (ConstantImpl<?>) application.function;
			application.breakIt(); // Freeing the constant and the argument
			return constant.apply(argument, this);
		}
		// Beta reduction
		AbstractionImpl abstraction = (AbstractionImpl) application.function;
		this.substitude(abstraction, abstraction.boundVariable.referenceCount,
				abstraction.boundVariable, argument);
		// The bound variable is no more referenced
		abstraction.boundVariable.breakIt();
		application.breakIt(); // Freeing the abstraction and the argument
		if (argument.referenceCount == 0)
			this.breakTermDeeply(argument);
		TermImpl body = abstraction.body;
		abstraction.breakIt(); // Freeing the body
		return body;
	}

	public void betaDeltaReduce(Term rootTerm, Application redex) {
		if (!redex.isRedex())
			throw new NotARedexException(rootTerm, redex);
		if (redex == rootTerm)
			throw new IllegalArgumentException(
					"Using this method, the redex cannot be the container");
		if (TermImpl.ASSERTION
				&& AssertUtils.referenceCountOf(rootTerm, redex) != redex
						.getReferenceCount())
			throw new RedexSharingException(rootTerm, redex);
		ArrayList<TermImpl> parentTermList = this.reservableTermList.reserve();
		ArrayList<Integer> argumentIndexList = this.reservableIntegerList
				.reserve();
		{
			this.locate((TermImpl) rootTerm, redex, redex.getReferenceCount(),
					parentTermList, argumentIndexList);
			replace(parentTermList, argumentIndexList, redex, null);
			TermImpl contraction = (TermImpl) this.betaDeltaReduce(redex);
			replace(parentTermList, argumentIndexList, null, contraction);
		}
		this.reservableTermList.makeBack();
		this.reservableIntegerList.makeBack();
	}

	public Application leftermostOutermostRedex(Term rootTerm, boolean isLazy) {
		if (rootTerm == null)
			throw new IllegalArgumentException();
		ArrayList<TermImpl> termStack = this.reservableTermStack.reserve();
		ArrayList<Integer> todoArgumentStack = this.reservableIntegerStack
				.reserve();
		HashSet<Term> explorationSet = this.reservableExplorationSet.reserve();
		TermImpl term = (TermImpl) rootTerm;
		int todoArgument = I_FUNCTION;
		try {
			while (true) {
				if (term.isVariable()
						|| (term.isConstant() && ((ConstantImpl<?>) term).argumentCount == 0)
						|| (isLazy && term.isAbstraction())
						|| (term.referenceCount > 1 && explorationSet
								.contains(term))) {
					while (true) {
						if (termStack.isEmpty())
							return null;
						int index = termStack.size() - 1;
						term = termStack.remove(index);
						todoArgument = todoArgumentStack.remove(index)
								.intValue();
						if (todoArgument != I_RETURN)
							break;
						if (term.referenceCount > 1)
							explorationSet.add(term);
					}
				} else if (term.isConstant()) {
					ConstantImpl<?> constant = (ConstantImpl<?>) term;
					if (todoArgument == constant.getArgumentCount() - 1) {
						if (term.referenceCount > 1) {
							termStack.add(term);
							todoArgumentStack.add(IND_RETURN);
						}
					} else {
						termStack.add(term);
						todoArgumentStack.add(toInteger(todoArgument + 1));
					}
					term = (TermImpl) constant.getArgument(todoArgument);
					if (term == null)
						throw new BrokenTermException(rootTerm, constant);
					todoArgument = I_FUNCTION;
				} else if (term.isAbstraction()) {
					if (term.referenceCount > 1) {
						termStack.add(term);
						todoArgumentStack.add(IND_RETURN);
					}
					AbstractionImpl abstraction = (AbstractionImpl) term;
					term = abstraction.body;
					if (term == null)
						throw new BrokenTermException(rootTerm, abstraction);
					todoArgument = I_FUNCTION;
				} else if (term.isApplication()) {
					ApplicationImpl application = (ApplicationImpl) term;
					if (todoArgument == I_ARGUMENT) {
						if (term.referenceCount > 1) {
							termStack.add(term);
							todoArgumentStack.add(IND_RETURN);
						}
						term = application.argument;
						if (term == null)
							throw new BrokenTermException(rootTerm, application);
						todoArgument = I_FUNCTION;
					} else {
						if (application.isRedex())
							return application;
						termStack.add(term);
						todoArgumentStack.add(IND_ARGUMENT);
						term = application.function;
						if (term == null)
							throw new BrokenTermException(rootTerm, application);
						todoArgument = I_FUNCTION;
					}
				} else
					throw new IllegalStateException();
			}
		} finally {
			this.reservableTermStack.makeBack();
			this.reservableIntegerStack.makeBack();
			this.reservableExplorationSet.makeBack();
		}
	}

	public Term normalForm(Term rootTerm, boolean isLazy,
			InformantReduction informantReduction) {
		if (((TermImpl) rootTerm).referenceCount > 0)
			throw new StillReferencedTermException(rootTerm);
		Application redex = this.leftermostOutermostRedex(rootTerm, isLazy);
		if (informantReduction != null)
			informantReduction.inform(InformantReduction.CHAR_REDUCTION_START,
					rootTerm, redex);
		while (redex != null) {
			char reductionType = redex.getFunction().isConstant() ? InformantReduction.CHAR_REDUCTION_DELTA
					: InformantReduction.CHAR_REDUCTION_BETA;
			try {
				if (redex == rootTerm)
					rootTerm = this.betaDeltaReduce(redex);
				else
					this.betaDeltaReduce(rootTerm, redex);
			} catch (TermException exception) {
				exception.setRootTerm(rootTerm);
				throw exception;
			}
			redex = this.leftermostOutermostRedex(rootTerm, isLazy);
			if (informantReduction != null)
				informantReduction.inform(reductionType, rootTerm, redex);
		}
		return rootTerm;
	}

	public Term normalForm(Term rootTerm) {
		return this.normalForm(rootTerm, false, null);
	}
}
