/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl;

/**
 * Statistic on Term use.
 */
public abstract class TermStatistic {

	// //////////////////////
	// Statistic on variables
	// //////////////////////

	/**
	 * Returns the count of free variables.
	 * 
	 * @return the count of free variables.
	 */
	public static final int getFreeVariableCount() {
		return VariableImpl.creationCount - VariableImpl.boundCount;
	}

	/**
	 * Returns the count of living bound variables.
	 * 
	 * @return the count of living bound variables.
	 */
	public static final int getLivingVariableCount() {
		return VariableImpl.creationCount - VariableImpl.breakCount;
	}

	/**
	 * Returns the count of consumed bound variables.
	 * 
	 * @return the count of consumed bound variables.
	 */
	public static final int getConsumedVariableCount() {
		return VariableImpl.breakCount;
	}

	/**
	 * Returns the stock number of reusable Variable.
	 * 
	 * @return the stock number of reusable Variable.
	 */
	public static final int getReusableVariableStockNumber() {
		return VariableImpl.reusableVariableStack.getSize();
	}

	// ///////////////////////////////////////
	// Statistic on constants with 2 arguments
	// ///////////////////////////////////////

	/**
	 * Returns the count of living constants with two arguments.
	 * 
	 * @return the count of living constants with two arguments.
	 */
	public static final int getLivingConstant2Count() {
		return Constant2Impl.creationCount - Constant2Impl.breakCount;
	}

	/**
	 * Returns the count of consumed constants with two arguments.
	 * 
	 * @return the count of consumed constants with two arguments.
	 */
	public static final int getConsumedConstant2Count() {
		return Constant2Impl.breakCount;
	}

	/**
	 * Returns the stock number of reusable constants with two arguments.
	 * 
	 * @return the stock number of reusable constants with two arguments.
	 */
	public static final int getReusableConstant2StockNumber() {
		return Constant2Impl.reusableConstantStack.getSize();
	}

	// ///////////////////////////////////////
	// Statistic on constants with 3 arguments
	// ///////////////////////////////////////

	/**
	 * Returns the count of living constants with three arguments.
	 * 
	 * @return the count of living constants with three arguments.
	 */
	public static final int getLivingConstant3Count() {
		return Constant3Impl.creationCount - Constant3Impl.breakCount;
	}

	/**
	 * Returns the count of consumed constants with three arguments.
	 * 
	 * @return the count of consumed constants with three arguments.
	 */
	public static final int getConsumedConstant3Count() {
		return Constant3Impl.breakCount;
	}

	/**
	 * Returns the stock number of reusable constants with three arguments.
	 * 
	 * @return the stock number of reusable constants with three arguments.
	 */
	public static final int getReusableConstant3StockNumber() {
		return Constant3Impl.reusableConstantStack.getSize();
	}

	// /////////////////////////
	// Statistic on abstractions
	// /////////////////////////

	/**
	 * Returns the count of living abstractions.
	 * 
	 * @return the count of living abstractions.
	 */
	public static final int getLivingAbstractionCount() {
		return AbstractionImpl.creationCount - AbstractionImpl.breakCount;
	}

	/**
	 * Returns the count of consumed abstractions.
	 * 
	 * @return the count of consumed abstractions.
	 */
	public static final int getConsumedAbstractionCount() {
		return AbstractionImpl.breakCount;
	}

	/**
	 * Returns the stock number of reusable Abstraction.
	 * 
	 * @return the stock number of reusable Abstraction.
	 */
	public static final int getReusableAbstractionStockNumber() {
		return AbstractionImpl.reusableAbstractionStack.getSize();
	}

	// /////////////////////////
	// Statistic on applications
	// /////////////////////////

	/**
	 * Returns the count of living applications.
	 * 
	 * @return the count of living applications.
	 */
	public static final int getLivingApplicationCount() {
		return ApplicationImpl.creationCount - ApplicationImpl.breakCount;
	}

	/**
	 * Returns the count of consumed applications.
	 * 
	 * @return the count of consumed applications.
	 */
	public static final int getConsumedApplicationCount() {
		return ApplicationImpl.breakCount;
	}

	/**
	 * Returns the stock number of reusable Application.
	 * 
	 * @return the stock number of reusable Application.
	 */
	public static final int getReusableApplicationStockNumber() {
		return ApplicationImpl.reusableApplicationStack.getSize();
	}
}
