/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.impl.type;

import java.util.List;

import org.xmloperator.lambda3.exception.type.ReplacedTermTypeException;
import org.xmloperator.lambda3.exception.type.UntypableTermException;
import org.xmloperator.lambda3.model.type.TermType;

/**
 * TermType variable implementation.
 * <p>
 * It is either terminal either replaced by another TermType.
 * </p>
 */
final class VariableTermTypeImpl extends TermTypeImpl implements TermType {
	TermType replacement = this;

	VariableTermTypeImpl() {
	}

	// ////////////////////////
	// TermType implementation
	// ////////////////////////

	public boolean isVariableTermType() {
		return true;
	}

	public boolean isTerminal() {
		return this.replacement == this;
	}

	public TermType getTerminalType() {
		if (this.replacement != this)
			this.replacement = this.replacement.getTerminalType();
		return this.replacement;
	}

	public void setEquivalence(TermType replacement) {
		if (this.replacement != this)
			throw new ReplacedTermTypeException();
		replacement = replacement.getTerminalType();
		if (replacement == this)
			return;
		if (replacement.isFunctionTermType() && replacement.isUsing(this))
			throw new UntypableTermException();
		this.replacement = replacement;
	}

	public boolean isUsing(TermType termType) {
		return (this == termType) ? true : (this.replacement == this) ? false
				: this.getTerminalType().isUsing(termType);
	}

	// //////////////////////
	// TermTypeImpl override
	// //////////////////////

	void toString(StringBuffer sb, List<TermType> terminalList) {
		TermType terminalType = this.getTerminalType();
		if (terminalType.isFunctionTermType())
			((FunctionTermTypeImpl) terminalType).toString(sb, terminalList);
		else if (terminalType.isConstantTermType())
			sb.append(((ConstantTermTypeImpl) terminalType).constantTypeChar);
		else {
			int index = terminalList.indexOf(terminalType);
			if (index < 0) {
				index = terminalList.size();
				terminalList.add(terminalType);
			}
			sb.append(TermType.CHAR_TYPE_NAME);
			sb.append(index + 1);
		}
	}
}
