/**
 * Copyright (c) 2013 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/lambda3/).
 */
package org.xmloperator.lambda3.util;

import org.xmloperator.lambda3.exception.AlreadyUsedResourceException;

/**
 * Abstract implementation of Reservable.
 * 
 * @param <T>
 *            the resource type.
 */
abstract class ReservableImpl<T> implements Reservable<T> {
	final T resource;
	boolean isReserved = false;
	
	ReservableImpl(T resource) {
		this.resource = resource;
	}
	
	/**
	 * Clears the resource.
	 */
	abstract void clearResource();
	
	////////////////////////////
	// Reservable implementation
	////////////////////////////

	public final T reserve() {
		if (this.isReserved)
			throw new AlreadyUsedResourceException();
		this.isReserved = true;
		return this.resource;
	}

	public final void makeBack() {
		if (!this.isReserved)
			throw new IllegalStateException();
		this.clearResource();
		this.isReserved = false;
	}
}
